/*************************************************************************/
/*                                                                       */
/* FILE NAME                                      VERSION                */
/*                                                                       */
/*      memory.c                   KS32C5000, KS32C50100   : version 1.0 */
/*                                                                       */
/* COMPONENT                                                             */
/*                                                                       */
/*                                                                       */
/* DESCRIPTION                                                           */
/*                                                                       */
/*                                                                       */
/* AUTHOR                                                                */
/*                                                                       */
/*                                                                       */
/* DATA STRUCTURES                                                       */
/*                                                                       */
/*                                                                       */
/* FUNCTIONS                                                             */
/*                                                                       */
/*      Evaluation & debugging utility for memory.                       */
/*                                                                       */
/* DEPENDENCIES                                                          */
/*                                                                       */
/*                                                                       */
/* HISTORY                                                               */
/*                                                                       */
/*         NAME            DATE                    REMARKS               */
/*                                                                       */
/*      hbahn           09-15-1998      Created initial version 1.0      */
/*                                                                       */
/*************************************************************************/

/* 			*/
/* Modified by 		*/
/* Dmitriy Cherkashin 	*/
/* dch@ucrouter.ru	*/
/* 2002			*/
/*			*/

//#include <stdlib.h>
#include "ks32c50.h"
#include "evm50100.h"


/////////////////////////////////////////////////////////////////////////////
// print content EXTDBWTH ///////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

static int PrintEXTDBWTH(char * nam, U32 val)
{
 int i;
// 
 if(nam == 0)
  {
   return(0);
  } 

// print register name & value //////////////////////////////////////////////

 Print("\n%12s %08x", nam, val);
 Print("\n%12s ", "");

// 
 for(i = 31; i >= 0; --i)
  {
   if( ((val >> i) & 0x01) != 0)
    {
     Print("1");
    }
   else
    {
     Print("0");
    }
  }
 Print("\n%12s 10987654321098765432109876543210", "");
 Print("\n%12s  3         2         1          ", "");

//
 Print("\nData bus width for:"); 
 Print("\n[ 1: 0] ROM/SRAM/FLASH bank 0 (DSR0)...[%x]", (val >> 0) & 3);
 Print("\n[ 3: 2] ROM/SRAM/FLASH bank 1 (DSR1)...[%x]", (val >> 2) & 3);
 Print("\n[ 5: 4] ROM/SRAM/FLASH bank 2 (DSR2)...[%x]", (val >> 4) & 3);
 Print("\n[ 7: 6] ROM/SRAM/FLASH bank 3 (DSR3)...[%x]", (val >> 6) & 3);
 Print("\n[ 9: 8] ROM/SRAM/FLASH bank 4 (DSR4)...[%x]", (val >> 8) & 3);
 Print("\n[10:11] ROM/SRAM/FLASH bank 5 (DSR5)...[%x]", (val >>10) & 3);
 Print("\n[13:12] DRAM bank 0 (DSD0).............[%x]", (val >>12) & 3);
 Print("\n[15:14] DRAM bank 1 (DSD1).............[%x]", (val >>14) & 3);
 Print("\n[17:16] DRAM bank 2 (DSD2).............[%x]", (val >>16) & 3);
 Print("\n[19:18] DRAM bank 3 (DSD3).............[%x]", (val >>18) & 3);
 Print("\n[21:20] external I/O bank 0 (DSX0).....[%x]", (val >>20) & 3);
 Print("\n[23:22] external I/O bank 1 (DSX1).....[%x]", (val >>22) & 3);
 Print("\n[25:24] external I/O bank 2 (DSX2).....[%x]", (val >>24) & 3);
 Print("\n[27:26] external I/O bank 3 (DSX3).....[%x]", (val >>26) & 3);

//
 Print("\n");
 return(1);
}

/////////////////////////////////////////////////////////////////////////////
// print ROMCON /////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

static int PrintROMCON(char * nam, U32 val)
{
 int i;
//
 if(nam == 0)
  {
   return(0);
  } 

//
 Print("\n%12s %08x", nam, val);
 Print("\n%12s ", "");
 for(i = 31; i >= 0; --i)
  {
   if( ((val >> i) & 0x01) != 0)
    {
     Print("1");
    }
   else
    {
     Print("0");
    }
  }
 Print("\n%12s 10987654321098765432109876543210", "");
 Print("\n%12s  3         2         1          ", "");

 Print("\n[1:0]   Page mode configuration (PMC)..[%x]", (val >> 0) & 3);
 switch( (val & ROMCON_PMC) )
  {
   case ROMCON_PMC_NORMAL  : Print("\n         00 = Normal ROM"); break;
   case ROMCON_PMC_4WPAGE  : Print("\n         01 = 4-word page"); break;
   case ROMCON_PMC_8WPAGE  : Print("\n         10 = 8-word page"); break;
   case ROMCON_PMC_16WPAGE : Print("\n         11 = 16-word page"); break;
   default :                 Print("\n         Internal error"); break;
  }

//
 Print("\n[3:2]   Page address access time (tPA).[%x]", (val >> 2) & 3);
 switch( (val & ROMCON_TPA) )
  {
   case ROMCON_TPA_5CLK    : Print("\n          00 = 5 cycles"); break;
   case ROMCON_TPA_2CLK    : Print("\n          01 = 2 cycles"); break;
   case ROMCON_TPA_3CLK    : Print("\n          10 = 3 cycles"); break;
   case ROMCON_TPA_4CLK    : Print("\n          11 = 4 cycles"); break;
   default                 : Print("\n          Internal error"); break; 
  }

// 
 Print("\n[6:4]   Programmable access cycle(tACC)[%x]", (val >> 4) & 7);
 switch( (val & ROMCON_TACC) )
  {
   case ROMCON_TACC_DISABLE: Print("\n          000 = Disable bank"); break;
   case ROMCON_TACC_2CLK   : Print("\n          001 = 2 cycles"); break;
   case ROMCON_TACC_3CLK   : Print("\n          010 = 3 cycles"); break;
   case ROMCON_TACC_4CLK   : Print("\n          011 = 4 cycles"); break;
   case ROMCON_TACC_5CLK   : Print("\n          100 = 5 cycles"); break;
   case ROMCON_TACC_6CLK   : Print("\n          101 = 6 cycles"); break;
   case ROMCON_TACC_7CLK   : Print("\n          110 = 7 cycles"); break;
   case ROMCON_TACC_UNUSED : Print("\n          111 = Reserved"); break;
   default :                 Print("\n          Internal error"); break;
  }

// 
 Print("\n[19:10]  ROM/SRAM/Flash base pointer...[%x]",  (val>>10) & 0x3FF);
 Print("\n         ..............................[%x]", ((val>>10) & 0x3FF) << 16);

 Print("\n[29:20]  ROM/SRAM/FLASH next pointer...[%x]",  (val>>20) & 0x3FF);
 Print("\n         ..............................[%x]", ((val>>20) & 0x3FF)<<16);

//
 Print("\n");
 return(1);
}

/////////////////////////////////////////////////////////////////////////////
// print content DRAMCON ////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////


static int PrintDRAMCON(char * nam, U32 val)
{
 int i;
//
 if(nam == 0)
  {
   return(0);
  } 

//
 Print("\n%12s %08x", nam, val);
 Print("\n%12s ", "");
//
 for(i = 31; i >= 0; --i)
  {
   if( ((val >> i) & 0x01) != 0)
    {
     Print("1");
    }
   else
    {
     Print("0");
    }
  }
 Print("\n%12s 10987654321098765432109876543210", "");
 Print("\n%12s  3         2         1          ", "");

////////////////////////////////////////////////////////////////////////////////
 Print("\n [0]      EDO mode(EDO) (note).........[%x]", (val >> 0) & 1);
 switch( (val & DRAMCON_EDO) )
  {
   case DRAMCON_NORMAL     : Print("\n          0 = Normal DRAM (Fast page mode DRAM)"); break;
   case DRAMCON_EDO        : Print("\n          1 = EDO DRAM"); break;
   default                 : Print("\n          Internal error"); break;
  }

////////////////////////////////////////////////////////////////////////////////
 Print("\n [2:1]    CAS strobe time (tCS)........[%x]", (val >> 1) & 3);
 switch( (val & DRAMCON_TCS) ) 
  {
   case DRAMCON_TCS_1CLK   : Print("\n          00 = 1 cycle"); break;
   case DRAMCON_TCS_2CLK   : Print("\n          01 = 2 cycles"); break;
   case DRAMCON_TCS_3CLK   : Print("\n          10 = 3 cycles"); break;
   case DRAMCON_TCS_4CLK   : Print("\n          11 = 4 cycles"); break;
   default                 : Print("\n          Internal error"); break;
  }

//////////////////////////////////////////////////////////////////////////////// 
 Print("\n [3:3]    CAS pre-charge time (tCP )...[%x]", (val >> 3) & 1);
 switch( (val & DRAMCON_TCP) )
  {
   case DRAMCON_TCP_1CLK   : Print("\n          0 = 1 cycle"); break;
   case DRAMCON_TCP_2CLK   : Print("\n          1 = 2 cycles"); break;
   default                 : Print("\n          Internal error"); break;
  }

// 
 Print("\n NOTE     In SDRAM mode, this bit affect SDRAM cycle.");
 Print("\n          tCS value [1] : 0 = 1 cycle, 1 = 2 cycle");

// 
 Print("\n [6:4]    Reserved ....................[%x]", (val >> 4) & 7);
 Print("\n          These bits default value is 000. But, you must set to 001.");

////////////////////////////////////////////////////////////////////////////////  
 Print("\n [7]      RAS to CAS delay(tRC or tRCD)[%x]", (val >> 7) & 1);
 switch( (val & DRAMCON_TRC) )
  {
   case DRAMCON_TRC_1CLK   : Print("\n           0 = 1 cycle"); break;
   case DRAMCON_TRC_2CLK   : Print("\n           1 = 2 cycles"); break;
   default                 : Print("\n           Internal error"); break;
  }

// 
 Print("\n [9:8]    RAS pre-charge time (tRP) ...[%x]", (val >> 8) & 3);
 switch( (val & DRAMCON_TRP) )
  {
   case DRAMCON_TRP_1CLK   : Print("\n           00 = 1 cycle"); break;
   case DRAMCON_TRP_2CLK   : Print("\n           01 = 2 cycles"); break;
   case DRAMCON_TRP_3CLK   : Print("\n           10 = 3 cycles"); break;
   case DRAMCON_TRP_4CLK   : Print("\n           11 = 4 cycles"); break;
  }

//
 Print("\n [19:10]  DRAM bank # base pointer ....[%x] [%x]",  
  (val>>10)&0x3FF		,
 ((val>>10)&0x3FF)<<16);

//
 Print("\n [29:20]  DRAM bank # Next pointer ....[%x] [%x]",  
  (val>>20)&0x3FF		,
 ((val>>20)&0x3FF)<<16);

//////////////////////////////////////////////////////////////////////////////// 
 Print("\n [31:30]  Number of column bits (CAN)..[%x]", (val >> 30) & 3);
 switch( (val & DRAMCON_CAN) )
  {
   case DRAMCON_CAN_8BITS  : Print("\n           00 = 8 bits"); break;
   case DRAMCON_CAN_9BITS  : Print("\n           01 = 9 bits"); break;
   case DRAMCON_CAN_10BITS : Print("\n           10 = 10 bits"); break;
   case DRAMCON_CAN_11BITS : Print("\n           11 = 11 bits"); break;
   default:                  Print("\n           Internal error"); break;
  }

//
 Print("\n");
 return(1);
}

/////////////////////////////////////////////////////////////////////////////
// Print REFEXTCON register /////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

static int PrintREFEXTCON(char * nam, U32 val)
{
 U32 prd;
 int i;
//
 if(nam == 0)
  {
   return(0);
  } 

//
 Print("\n%12s %08x", nam, val);	// print register name & value   
 Print("\n%12s ", "");			// 
//
 for(i = 31; i >= 0; --i)
  {
   Print
    (
     "%s"		, 		// print register bits
    ((val >> i) & 0x01)	?		
     "1"		:
     "0"
    );
  }
 Print("\n%12s 10987654321098765432109876543210", "");
 Print("\n%12s  3         2         1          ", "");

// 
 Print("\n[9:0]     External I/O base address ...[%x]", ((val>>0) & 0x3FF));
 Print("\n          .............................[%x]", ((val>>0) & 0x3FF)<<16);

////////////////////////////////////////////////////////////////////////////////
 Print("\n[15]      Validity of special reg(VSF).%s"	,
   (val & REFEXTCON_VSF) 		?
  "1(Accessible to memory bank)" 	:
  "0(Not accessible to memory bank)"
  );

////////////////////////////////////////////////////////////////////////////////
 Print("\n[16]      Refresh enable (REN) ........%s"	,
   (val & REFEXTCON_REN) 		?	
   "1(Enable DRAM refresh)"		:
   "0(Disable DRAM refresh)"
  );

//   
 Print("\n[19:17]   CAS hold time(tCHR)..........");
 switch( (val & REFEXTCON_TRC) )
  {
   case REFEXTCON_TRC_1CLK : Print("000 = 1 cycle");  break;
   case REFEXTCON_TRC_2CLK : Print("001 = 2 cycles"); break;
   case REFEXTCON_TRC_3CLK : Print("010 = 3 cycles"); break;
   case REFEXTCON_TRC_4CLK : Print("011 = 4 cycles"); break;
   case REFEXTCON_TRC_5CLK : Print("100 = 5 cycles"); break;
   case REFEXTCON_TRC_6CLK : Print("101 = Not used (6 cycles)"); break;
   default: Print("\nNot used"); break;
  }

// 
 Print("\n NOTE :    In EDO/normal DRAM mode, CAS hold time can be programmed upto");
 Print("\n           5 cycles. But in SDRAM mode, this bit fields function are");
 Print("\n           defined as ROW Cycle Time (tRC) and can be programmed upto");
 Print("\n           6 cycles.");

////////////////////////////////////////////////////////////////////////////////
 Print("\n[20]      CAS setup time(t CSR)........"	,
  (val & REFEXTCON_TCSR) 	?
   "1(2 cycles)"		:
   "0(1 cycle)"
  );

 Print("\nNOTE :    In SDRAM mode, this bit field is reserved.");

//
 Print("\n[31:21]   Refresh count value (duration) ...[%x]", ((val >> 21)));
 Print("\n          The refresh period is calculated as (2 ** 11 - Value + 1) / fMCLK"); 

//
 prd = ( (2048  - (val >> 21) + 1) * 1000 * 1000) / 50000000;  
 Print("\n[31:21]   Refresh period (us) ..............[%d]", prd);

//
 Print("\n");
 return(1);
}

/////////////////////////////////////////////////////////////////////////////
// print EXTACON0/EXTACON1 //////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////


static int PrintEXTACON(int num)
{
 U32 val;				// register value 
// 
 if(num == 0)
  {
   val = EXTACON0;			// EXTACON0 
  }
 else
 if(num == 1) 
  {
   val = EXTACON1;			// EXTACON1 
  }
 else 
  {
   return(0);
  } 


/////////////////////////////////////////////////////////////////////////////
 Print("\nEXTCON%d 0x%08x", num, val);

/* [0:2]   EXTACON0 tCOS0 */
/* [0:2]   EXTACON1 tCOS2 */

 Print("\n[0:2]   Chip selection set-up time on nOE %s=%d"	,
    num 	? 
  "tCOS2"	:
  "tCOS0"	,
  (val & 0x7) 
  );

/* [5:3]   EXTACON0 tACS0 */
/* [5:3]   EXTACON1 tACS2 */

 Print("\n[5:3]   Address set-up time before nECS %s=%d" 	,
   num		?
  "tACS2"	:
  "tACS0"	,
  ((val >> 3) & 0x7) 
  );

/* [8:6]   EXTACON0 tCOH0 */
/* [8:6]   EXTACON1 tCOH2 */

 Print("\n[8:6]   Chip selection hold time on nOE %s=%d"	,
   num		?
   "tCOH2"	:
   "tCOH0"	,
   ((val >> 6) & 0x7)
  );

/* [11:9]  EXTACON0 tACC0 */
/* [11:9]  EXTACON1 tACC2 */

 Print("\n[11:9]  Access cycles nOE low time      %s=%d"	,
   num		?
   "tACC2"	:
   "tACC0"	,
   ((val >> 9) & 0x7) 
  );


/* [18:16] EXTACON0 tCOS1 */
/* [18:16] EXTACON1 tCOS3 */

 Print("\n[18:16] Chip selection set-up time on nOE %s=%d"	,
   num		?
  "tCOS3"	:
  "tCOS1"	,
  ((val >> 16) & 0x7)
 );


/* [21:19] EXTACON0 tACS1 */
/* [21:19] EXTACON1 tACS3 */

 Print("\n[21:19] Address set-up time before nECS   %s=%d"	,
   num		?
  "tACS3"	:
  "tACS1"	,
  ((val >> 19) & 0x7) 
 );


/* [24:22] EXTACON0 tCOH1 */
/* [24:22] EXTACON1 tCOH3 */

 Print("\n[24:22] Chip selection hold time on nOE  %s=%d"	,
   num		?
  "tCOH3"	:
  "tCOH1"	,
  ((val >> 22)& 0x7)
 );


/* [27:25] EXTACON0 tACC1 */
/* [27:25] EXTACON1 tACC3 */

 Print("\n[27:25] Access cycles nOE low time       %s=%d"	,
  num		?
  "tACC3"	:
  "tACC1"	,
  ((val >> 25)& 0x7)
 );

// 
 Print("\n");
 return(1);
}

/////////////////////////////////////////////////////////////////////////////
// Print Memory Configuration ///////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

static int MemViewConf(void)
{
 while(1)
  {
   U8 it;
   Print("\n\nROM/DRAM Configuration");  
   Print("\n[W] EXTDBWTH External data bus width register.");
   Print("\n[0] ROMCON0  ROM/SRAM/FLASH bank 0 register.");
   Print("\n[1] ROMCON1  ROM/SRAM/FLASH bank 1 register.");
   Print("\n[2] ROMCON2  ROM/SRAM/FLASH bank 2 register.");
   Print("\n[3] ROMCON3  ROM/SRAM/FLASH bank 3 register.");
   Print("\n[4] ROMCON4  ROM/SRAM/FLASH bank 4 register.");
   Print("\n[5] ROMCON5  ROM/SRAM/FLASH bank 5 register.");
   Print("\n[A] DRAMCON0 DRAM bank 0 control register.");
   Print("\n[B] DRAMCON1 DRAM bank 1 control register.");
   Print("\n[C] DRAMCON2 DRAM bank 2 control register.");
   Print("\n[D] DRAMCON3 DRAM bank 3 control register.");
   Print("\n[R] REFEXTCON.");
   Print("\n[E] EXTACON0.");
   Print("\n[X] EXTACON1.");
   Print("\n[Q] Quit.");
   Print("\n    Select test item ");
   it = get_upper();
   switch(it)
    {
     case 'W' :  
// EXTDBWTH External data bus width register ///////////////////////////////////
      PrintEXTDBWTH("EXTDBWTH", EXTDBWTH);	//
      break;

     case '0' :
// ROMCON0  ROM/SRAM/FLASH bank 0 register /////////////////////////////////////
      PrintROMCON("ROMCON0", ROMCON0);		//
      break;
      
     case '1' : 
// ROMCON1  ROM/SRAM/FLASH bank 1 register /////////////////////////////////////
      PrintROMCON("ROMCON1", ROMCON1);		//
      break;

     case '2' :
// ROMCON2  ROM/SRAM/FLASH bank 2 register /////////////////////////////////////
      PrintROMCON("ROMCON2", ROMCON2);		//
      break;
      
     case '3' :
// ROMCON3  ROM/SRAM/FLASH bank 3 register /////////////////////////////////////
      PrintROMCON("ROMCON3", ROMCON3);		//
      break;
      
     case '4' :
// ROMCON4  ROM/SRAM/FLASH bank 4 register /////////////////////////////////////
      PrintROMCON("ROMCON4", ROMCON4);		//
      break;
        
     case '5' : 
// ROMCON5  ROM/SRAM/FLASH bank 5 register /////////////////////////////////////
      PrintROMCON("ROMCON5", ROMCON5); 		//
      break;
      
     case 'A' :
// DRAMCON0 DRAM bank 0 control register ///////////////////////////////////////
      PrintDRAMCON("DRAMCON0",DRAMCON0);	//	
      break;
      
     case 'B' :
// DRAMCON0 DRAM bank 1 control register ///////////////////////////////////////
      PrintDRAMCON("DRAMCON1",DRAMCON1);	//
      break;
      
     case 'C' :
// DRAMCON0 DRAM bank 2 control register ///////////////////////////////////////
      PrintDRAMCON("DRAMCON2",DRAMCON2);	//
      break;
      
     case 'D' :
// DRAMCON0 DRAM bank 3 control register ///////////////////////////////////////
      PrintDRAMCON("DRAMCON3",DRAMCON3);	//
      break;
      
     case 'R' :
// REFEXTCON /////////////////////////////////////////////////////////////////// 
      PrintREFEXTCON("REFEXTCON", REFEXTCON);	//
      break; 

     case 'E' :
// EXTACON0 //////////////////////////////////////////////////////////////////// 
      PrintEXTACON(0);				//
      break;

     case 'X' :
// EXTACON1 //////////////////////////////////////////////////////////////////// 
      PrintEXTACON(1);				//
      break;
// 
     case 'Q' :
      return(1);
// 
     default  :
      break;
    } 

   Print("\nPress any key to continue");
   it = get_byte();
   if(it == 'Q' || it == 'q')
    {
     return(1);
    } 
  }
}

// source constant /////////////////////////////////////////////////////////////

#define MemAddrCheck_DRAMCON0 0x0001	// DRAMCON0 (DRAM bank 0) 
#define MemAddrCheck_DRAMCON1 0x0002	// DRAMCON1 (DRAM bank 1)
#define MemAddrCheck_DRAMCON2 0x0004	// DRAMCON2 (DRAM bank 2)
#define MemAddrCheck_DRAMCON3 0x0008	// DRAMCON3 (DRAM bank 3)

#define MemAddrCheck_ROMCON0  0x0010	// ROMCON0 (ROM/SRAM/FLASH bank 0)  
#define MemAddrCheck_ROMCON1  0x0020 	// ROMCON1 (ROM/SRAM/FLASH bank 1)
#define MemAddrCheck_ROMCON2  0x0040 	// ROMCON2 (ROM/SRAM/FLASH bank 2)
#define MemAddrCheck_ROMCON3  0x0080 	// ROMCON3 (ROM/SRAM/FLASH bank 3) 
#define MemAddrCheck_ROMCON4  0x0100 	// ROMCON4 (ROM/SRAM/FLASH bank 4)
#define MemAddrCheck_ROMCON5  0x0200 	// ROMCON5 (ROM/SRAM/FLASH bank 5) 
//

int MemRegions(U32 * base, U32 * limit, U32 * src, int *size, int dramonly)
{
 int cnt = 0;				// 
 int i;				
 int j;

 if(base == 0 || limit == 0 || src == 0 || size == 0)
  {
   return(0);
  } 

//
 if(dramonly == 0)
  {
/* [19:10]  ROM/SRAM/Flash bank # base pointer				      */
/*	    This value is the start address of the ROM/SRAM/Flash bank #.     */
/*          The start address is calculated as ROM/SRAM/FLASH bank # base     */
/*          pointer << 16 						      */
/* 									      */	
/* [29:20]  ROM/SRAM/FLASH bank # next pointer 			              */
/*          This value is the current bank end address << 16 + 1 	      */

   if((EXTDBWTH & EXTDBWTH_DSR0) != 0)
    { 
// Data bus width for ROM/SRAM/FLASH bank 0 ////////////////////////////////////

     if(cnt < *size)
      {
       *(base  + cnt) = (ROMCON0 >> 10) & 0x3FF;// base address of ROM bank 0  
       *(limit + cnt) = (ROMCON0 >> 20) & 0x3FF;// next address for ROM bank 0  
       *(src + cnt++) = MemAddrCheck_ROMCON0;	// source = ROMCON0
      }
    }

   if((EXTDBWTH & EXTDBWTH_DSR1) != 0)
    {
// Data bus width for ROM/SRAM/FLASH bank 1 ////////////////////////////////////
     if(cnt < *size)
      {
       *(base  + cnt) = (ROMCON1 >> 10) & 0x3FF;// base address of ROM bank 1
       *(limit + cnt) = (ROMCON1 >> 20) & 0x3FF;// next address for ROM bank 1
       *(src + cnt++) = MemAddrCheck_ROMCON1;	// source = ROMCON1 
      } 
    }

   if((EXTDBWTH & EXTDBWTH_DSR2) != 0) 
    {
// Data bus width for ROM/SRAM/FLASH bank 2 ////////////////////////////////////
     if(cnt < *size)
      {
       *(base  + cnt) = (ROMCON2 >> 10) & 0x3FF;// base address of ROM bank 2 
       *(limit + cnt) = (ROMCON2 >> 20) & 0x3FF;// next address for ROM bank 2
       *(src + cnt++) = MemAddrCheck_ROMCON2;	// source = ROMCON2 
      }
    }

   if((EXTDBWTH & EXTDBWTH_DSR3) != 0) 
    {
// Data bus width for ROM/SRAM/FLASH bank 3 ////////////////////////////////////
     if(cnt < *size)
      { 
       *(base  + cnt) = (ROMCON3 >> 10) & 0x3FF;// base address of ROM bank 3
       *(limit + cnt) = (ROMCON3 >> 20) & 0x3FF;// next address for ROM bank 3
       *(src + cnt++) = MemAddrCheck_ROMCON3;	// source = ROMCON3
      }
    }
  
   if((EXTDBWTH & EXTDBWTH_DSR4) != 0) 
    {
// Data bus width for ROM/SRAM/FLASH bank 4 ////////////////////////////////////
     if(cnt < *size)
      {
       *(base  + cnt) = (ROMCON4 >> 10) & 0x3FF;// base address of ROM bank 4 	
       *(limit + cnt) = (ROMCON4 >> 20) & 0x3FF;// next address for ROM bank 4
       *(src + cnt++) = MemAddrCheck_ROMCON4;	// source = ROMCON4
      }
    }
  
   if((EXTDBWTH & EXTDBWTH_DSR5) != 0) 
    {
// Data bus width for ROM/SRAM/FLASH bank 5 ////////////////////////////////////
     if(cnt < *size)
      {
       *(base  + cnt) = (ROMCON5 >> 10) & 0x3FF;// base address of ROM bank 5
       *(limit + cnt) = (ROMCON5 >> 20) & 0x3FF;// next address for ROM bank 5
       *(src + cnt++) = MemAddrCheck_ROMCON5;	// source = ROMCON5 
      }
    }
  }

/* [19:10]  DRAM bank # base pointer 					      */
/*          This value indicates the start address of DRAM bank #. 	      */
/*          The start address is calculated as RAM bank # base pointer << 16  */
/* 									      */
/* [29:20]  DRAM bank # Next pointer 					      */
/*          This value isthe current bank end address << 16 + 1               */


 if((EXTDBWTH & EXTDBWTH_DSD0) != 0) 
  {
// Data bus width for DRAM bank 0 //////////////////////////////////////////////
   if(cnt < *size)
    {
     *(base  + cnt) = (DRAMCON0 >> 10) & 0x3FF;	// DRAM bank 0 base address
     *(limit + cnt) = (DRAMCON0 >> 20) & 0x3FF;	// DRAM bank 0 next address	
     *(src + cnt++) = MemAddrCheck_DRAMCON0;	// source = DRAMCON0
    } 
  }

 if((EXTDBWTH & EXTDBWTH_DSD1) != 0) 
  {
// Data bus width for DRAM bank 1 //////////////////////////////////////////////
   if(cnt < *size)
    {
     *(base  + cnt) = (DRAMCON1 >> 10) & 0x3FF;	// DRAM bank 1 base address
     *(limit + cnt) = (DRAMCON1 >> 20) & 0x3FF;	// DRAM bank 1 next address
     *(src + cnt++) = MemAddrCheck_DRAMCON1;	// source = DRAMCON1
    }
  }
  
 if((EXTDBWTH & EXTDBWTH_DSD2) != 0) 
  {
// Data bus width for DRAM bank 2 //////////////////////////////////////////////
   if(cnt < *size)
    {
     *(base  + cnt) = (DRAMCON2 >> 10) & 0x3FF;	// DRAM bank 2 base address
     *(limit + cnt) = (DRAMCON2 >> 20) & 0x3FF;	// DRAM bank 2 next address
     *(src + cnt++) = MemAddrCheck_DRAMCON2;	// source = DRAMCON2
    }
  }

 if((EXTDBWTH & EXTDBWTH_DSD3) != 0) 
  {
// Data bus width for DRAM bank 3 //////////////////////////////////////////////
   if(cnt < *size)
    {
     *(base  + cnt) = (DRAMCON3 >> 10) & 0x3FF;	// DRAM bank 3 base address
     *(limit + cnt) = (DRAMCON3 >> 20) & 0x3FF;	// DRAM bank 3 next address
     *(src + cnt++) = MemAddrCheck_DRAMCON3;	// source = DRAMCON3
    }
  }

// sort base/limit/str /////////////////////////////////////////////////////////

 for(i = 0; i + 1 < cnt; ++i)
  {
   for(j = i + 1; j < cnt; ++j)
    {
     if(*(base + j) < *(base + i))
      {
       U32 tmp;
//
       tmp = *(base + j);
       *(base + j) = *(base + i);
       *(base + i) = tmp;

//
       tmp = *(limit + j);
       *(limit + j) = *(limit + i);
       *(limit + i) = tmp;

//
       tmp = *(src + j);
       *(src + j) = *(src + i);
       *(src + i) = tmp;
      }
    }
  }

// 
 j = 0;
 i = 0;
 while(i < cnt)
  {
   *(base  + j) = *(base + i);
   *(limit + j) = *(limit + i);
   *(src + j) = *(src + i);
   i++;

   while(i < cnt)
    {
     if(*(limit + j) == *(base + i))
      {
       *(limit + j) = *(limit + i);
       *(src + j)  |= *(src + i);
      }
     else
      {
       break;
      }
//
     ++i;
    }
//
   ++j;
  }   

//
 cnt = j;

//
 for(i = 0; i < cnt; ++i)
  {
   *(base  + i) <<= 16;
   *(limit + i) <<= 16;
  }

//
 *size = cnt;
 return(1);
} 
 

/////////////////////////////////////////////////////////////////////////////////
// return free dram regions /////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////

int FreeDramRegions(U32 * base, U32 * limit, int *size)
{
 U32 src[12];				// sourse 
 U32 ubase[3];				// used region base 
 U32 ulimit[3];				// used region limit 
 int cnt;				// count of regions 
 int un;				// used region number
 int n;

 if(base == 0 || limit == 0 || size == 0)
  {
   return(0);
  } 

 if(*size > 12)
  {					// src
   *size = 12;
  } 

// 
 cnt = *size;
 if(MemRegions(base, limit, src, & cnt, 1) == 0)
  {
   return(0);
  } 

// 
 ubase [0] = (U32)Image_RO_Base;	// Image Read Only Base
 ulimit[0] = (U32)Image_RO_Limit;	// Image Read Only Limit
 ubase [1] = (U32)Image_RW_Base;	// Image Read/Write Base
 ulimit[1] = (U32)Image_RW_Limit;	// Image Read/Write Limit
 ubase [2] = (U32)Image_ZI_Base;	// Image Zero Init Base
 ulimit[2] = (U32)Image_ZI_Limit;	// Image Zero Init Limit

 for(un = 0; un < 3; ++un)
  {
   if(ulimit[un] <= ubase[un])
    {					// not used 
     continue;
    } 
// 
   for(n = 0; n < cnt; ++n)
    {
     if(limit[n] <= base[n])
      {					// not used 
       continue;
      } 

     if(ubase[un] >= limit[n] || ulimit[un] <= base[n])
      {
// .....uuuuuuuu.O.uuuuuu 
// ffffff........R......fffffff
       continue;
      } 

     if(ubase[un] <= base[n])
      {
// uuuuuuuuuu
//   fffffff
//   fffffffffff 
       if(limit[n] <= ulimit[un])
        {				
         base[n] = limit[n];		// remove 
        }
       else
        {
         base[n] = ulimit[un];
        }
      }
     else 
      {
//  uuuuuuuuu//
//ffffff
//fffffffffffff 
       if(limit[n] <= ulimit[un])
        {
         limit[n] = ubase[un];		// 
        }
       else
        {
         if(cnt < *size)
          {
           base [cnt] = ulimit[un];
           limit[cnt] = limit[n];
           limit[n]   = ubase[un];
           cnt++;
          }
         else
          {
           return(0);
          }  
        }
      }
    }    
  }

//
// 
 un = 0;
 n = 0;
 while(n < cnt)
  {
   if(*(base + n) < *(limit + n))
    {
     *(base  + un) = *(base  + n);
     *(limit + un) = *(limit + n);
      un++; 
    } 
   n++;
  }   

//
 *size = un;
  return(1);
}


////////////////////////////////////////////////////////////////////////////////
// check memory region /////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

//
// saddr    = start address
// eaddr    = end address (first byte that not included int tested memory region) 
// dramonly = 

static int MemAddrCheck(U32 * saddr, U32 * eaddr, int dramonly) 
{
 U32 start;				// test region start address
 U32 end; 				// test region end address
// 
 U32 base[10];				// ROM/RAM bank base address 
 U32 limit[10];				// ROM/RAM bank next address
 U32 src[10];				// base/limit source
 int cnt;
 int i;

 cnt = 10;
 if(MemRegions(base, limit, src, & cnt, dramonly) == 0) 
  {
   return(0);
  }

// 
 start = (U32) (saddr);			// tested region start address
 end   = (U32) (eaddr); 		// tested region end address
 
 for(i = 0; i < cnt; ++i)
  {
   if(start >= base[i] && end <= limit[i])
    {
// start & end address in continuous memory region /////////////////////////////
     return(1);
    } 
  }
 
//
 Print
  (
  "\nMemory from 0x%08x to 0x%08x out of configured memory" ,
     saddr, 
     eaddr
  ); 


// print configured memory regions /////////////////////////////////////////////
 for(i = 0; i < cnt; ++i)
  {
   Print("\n  0x%08x 0x%08x", base[i], limit[i]);

   if((src[i] & MemAddrCheck_ROMCON0 ) != 0) { Print("  ROMCON0"); }
   if((src[i] & MemAddrCheck_ROMCON1 ) != 0) { Print("  ROMCON1"); }
   if((src[i] & MemAddrCheck_ROMCON2 ) != 0) { Print("  ROMCON2"); }
   if((src[i] & MemAddrCheck_ROMCON3 ) != 0) { Print("  ROMCON3"); }
   if((src[i] & MemAddrCheck_ROMCON4 ) != 0) { Print("  ROMCON4"); }
   if((src[i] & MemAddrCheck_ROMCON5 ) != 0) { Print("  ROMCON5"); }

   if((src[i] & MemAddrCheck_DRAMCON0) != 0) { Print(" DRAMCON0"); }
   if((src[i] & MemAddrCheck_DRAMCON1) != 0) { Print(" DRAMCON1"); }
   if((src[i] & MemAddrCheck_DRAMCON2) != 0) { Print(" DRAMCON2"); }
   if((src[i] & MemAddrCheck_DRAMCON3) != 0) { Print(" DRAMCON3"); }
  }

// 
 Print("\n");
 return(0); 
}




////////////////////////////////////////////////////////////////////////////////
// Memory dump program /////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////


void MemDump(U32 * base, U32 * end)
{
 int cnt;				// words count
 int row;				// current row 
 int i;					// start word in row 
 int j;					// word index in row 

// 
 cnt = end - base;			// words count 
 row = 0; 				// init current row 
 j = 0;					// 

 for(i=0; i < cnt + 4; i+= j)
  {
   if(row == 20) 
    {
     U8 it;
     Print("\nPress Any Key"); 
     it = get_byte();
     if(it == 'Q' || it == 'q')
      {
       return;
      } 
     row = 0;
    }
   else 
    { 
     row++;
    
     Print("\n%08x : ", base);
     for(j = 0; j < 4 ; ++j)
      {
// 
       U8 * ptr;
       int bn;
       ptr = (U8 *) base;
       for(bn = 0; bn < 4; ++bn)
        { 
         Print("%02x", *ptr);
         ptr++;
        } 
// 
       Print(" ");
       base++;
      }
    }
  } 
}

////////////////////////////////////////////////////////////////////////////////
// Memory Dump top function ////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////


void MemoryDump(void)
{
 U32   dramcon0;			// Dram bank 0 address from DRAMCON0
 U32   addr;				// dump memory address
 U32   size;				// dump memory size
 U32 * base;  				// Dump Memory base address
 U32 * end;   				// Dump Memory end address

/* [19:10]  DRAM bank # base pointer 					      */
/*          This value indicates the start address of DRAM bank #. 	      */
/*          The start address is calculated as RAM bank # base pointer << 16  */

 dramcon0 = ( (DRAMCON0 >> 10) & 0x3FF ) << 16;

//    
 Print("\nInput Memory Dump Base Address (0x%08x) : 0x", dramcon0);	
 addr = get_number(16,0);
 if(addr == 0) 
  {					/* default start dump address */  
   addr = dramcon0;
  }

 Print("\nInput Memory Dump Size (0x100)              : 0x") ;	
 size = get_number(16,0);
 if(size == 0) 
  {
   size=0x100;
  }

// 
 base = (U32 *) addr;			// base address
 end  = (U32 *)(addr + size);		// end address

 if(MemAddrCheck(base,end,0) == 0)
  {
// memory Region out of configured memory //////////////////////////////////////
   U8 ch;
   Print("\nContinue dump [Y/N]_");
   ch = get_upper(); 
   if(ch != 'Y')
    { 
     return;
    } 
  }


 Print("\nMemory Dump from %x to %x",base, end) ;
 MemDump(base, end);
}

////////////////////////////////////////////////////////////////////////////////
// Memory Pattern Fill /////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

static void MemPatternFill(void) 
{
 U32 * start;				// start address 
 U32 * end;				// end address
 U32 pat;				// pattern 
 int cnt;				// words count 
 U8 ch;	

// 
 Print("\nMemory Pattern Fill");
 Print("\nMemory Pattern Fill Start Address : 0x");
 start = (U32 *) get_number(16,0);

//  
 Print("\nMemory Pattern Fill End Address   : 0x");
 end   = (U32 *) get_number(16,0);

//  
 Print("\nMemor Fill Pattern (word)         : 0x");
 pat   = get_number(16,0);

 cnt = (int)(end - start);
 if(cnt < 0)
  {					// start address >
   return;				// end address
  } 

 if(MemAddrCheck(start,end,0) == 0)
  {
// memory Region out of configured memory //////////////////////////////////////
   Print
    (
     "\nMemory region 0x%08x 0x%08x out of configured memory"	,
       start	,			// start address 
       end				// end address
    );
  }


 Print
  (
   "\nMemory Fill %08x to %08x, Pattern : %08x [Y/N]_",
//
    start	, 			//
    end		, 			//
    pat					// 
  );

 ch = get_upper();
 if(ch == 'Y')
  {
   while( cnt-- ) 
    {
     *start++ = pat;
    }
  }
}


////////////////////////////////////////////////////////////////////////////////
// Memory Pattern Search ///////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////


static void MemPatternSearch(void)
{
 U32 * start;
 U32 * end;
 U32 pat;
 int cnt;

 Print("\nMemory Pattern Search");
 Print("\nMemory Pattern Search Start Address : 0x");	
 start = (U32 *)get_number(16,0);
 
 Print("\nMemory Pattern Search End Address   : 0x");	
 end = (U32 *)get_number(16,0);
 
 Print("\nMemory Search Pattern (word)        : 0x");	
 pat = get_number(16,0);

 cnt = (int)( end - start );


 if(MemAddrCheck(start,end,0) == 0)
  {
   U8 ch;
   Print("\nContinue Search [Y/N]_");
   ch = get_upper(); 
   if(ch != 'Y')
    {
     return;
    } 
  }

 Print
  (
   "\nMemory Pattern Search 0x%08x to 0x%08x, Pattern : %08x\n",
    start	, 
    end		,
    pat
  );

 while(cnt--) 
  {
   if( *start == pat ) 
    {
     Print
      (
       "\nPattern is Searched : 0x%08x (0x%08x)",
         start	, 			// address 
        *start				// pattern 
      );

// 
     if(kbd_hit())
      {
       U8 it;
       it = get_byte();
       if(it == KEY_ESC)
        {
         return;
        } 
      }
    }
   start++;
  }
}


////////////////////////////////////////////////////////////////////////////////
// Byte Copy Function //////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////


int bcopy(void *src, void *dst, int words, int hash)
{
 U8 *SrcAddr; 				// 
 U8 *DstAddr;				// 
 int cnt;				// 
 int hnum;				// progress indicator

 SrcAddr = (U8 *) src;			/* byte pointer to source */
 DstAddr = (U8 *) dst;			/* byte pointer to destination */
 cnt	 = words * 4;			/* byte count */

 if(hash <= 0)
  {
// no print progress ///////////////////////////////////////////////////////////
   while(cnt--)
    {
     *DstAddr++ = *SrcAddr++ ;
    }
//
   return(1);
  }

// 
 hnum = 0;
 while(cnt > 0)
  {
   int hcnt;
   if(cnt > hash) { hcnt = hash; }
   else           { hcnt = cnt;  }    
   cnt -= hcnt;				// remain copy bytes

   while(hcnt--)
    {
     *DstAddr++ = *SrcAddr++;
    }
//
   PrintRotSlash(hnum++);		// print progress indicator
   if(kbd_hit())
    {
     U8 it;
     it = get_byte();
     if(it == KEY_ESC)
      {
       return(-1);			// bcopy skiped
      } 
    }
  }

//
 return(1);
}

////////////////////////////////////////////////////////////////////////////////
// Half Word Copy Function /////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////


int scopy(void *src, void *dst, int words, int hash)
{
 U16 *SrcAddr; 					
 U16 *DstAddr;
 int cnt;
 int hnum;

 SrcAddr = (U16 *) src;			/* half word pointer to source */
 DstAddr = (U16 *) dst;			/* half word pointer to destination */
 cnt	 = words * 2;			/* half word count */

 if(hash <= 0)
  {
// no print half word copy progress ////////////////////////////////////////////
   while(cnt--)
    {
     *DstAddr++ = *SrcAddr++;
    }
//
   return(1);
  }  

// 
 hnum = 0;
 while(cnt > 0)
  {
   int hcnt;				//
   if(cnt > hash) { hcnt = hash; }
   else           { hcnt = cnt; }    
   cnt -= hcnt;				// remain half word copy

//    
   while(hcnt--)
    {
     *DstAddr++ = *SrcAddr++;
    }
//
   PrintRotSlash(hnum++);		// print progress indicator
   if(kbd_hit())
    {
     U8 it;
     it = get_byte();
     if(it == KEY_ESC)
      {
       return(-1);			// bcopy skiped
      } 
    }
  }  
// 
 return(1);
}

////////////////////////////////////////////////////////////////////////////////
// Word Copy Function //////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////


int wcopy(void *src, void *dst, int words, int hash)
{
 U32 *SrcAddr; 
 U32 *DstAddr;
 int cnt;
 int hnum;				// progress indicator

 SrcAddr = (U32 *)src;			/* word pointer to source */
 DstAddr = (U32 *)dst;			/* word pointer to destination */
 cnt	 = words;			/* word count */

 if(hash <= 0)
  {
// no print progress indicator /////////////////////////////////////////////////
   while(cnt--)
    {
     *DstAddr++ = *SrcAddr++;
    }
// 
   return(1); 
  }  


 hnum = 0;
 while(cnt > 0)
  {
   int hcnt;				//
   if(cnt > hash) { hcnt = hash; }
   else           { hcnt = cnt; }    
   cnt -= hcnt;				// remain half word copy
   
   while(hcnt--)
    {
     *DstAddr++ = *SrcAddr++;
    }
//
   PrintRotSlash(hnum++);		// print progress indicator
   if(kbd_hit())
    {
     U8 it;
     it = get_byte();
     if(it == KEY_ESC)
      {
       return(-1);			// bcopy skiped
      } 
    }
  }
// 
 return(1);  
}

////////////////////////////////////////////////////////////////////////////////
// Byte Compare Function ///////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////


int bcomp(void *src, void *dst, int words, int hash)
{
 U8 * SrcAddr; 
 U8 * DstAddr;
 int cnt;
 int hnum;				// progress indicator

 SrcAddr = (U8 *) src;			/* byte pointer to source */
 DstAddr = (U8 *) dst;			/* byte pointer to destination */
 cnt	 = words * 4;			/* byte count */

 if(hash <= 0)
  {
// no print progress indicator /////////////////////////////////////////////////

   while(cnt--)
    {
     if(*DstAddr != *SrcAddr)
      {
       Print("\nMemory Byte Test Error") ;
       Print("\nSource Data is %02x at Addr %08x",*SrcAddr,SrcAddr);
       Print("\nDestin Data is %02x at Addr %08x",*DstAddr,DstAddr);
       return(0);
      }
     DstAddr++;
     SrcAddr++;
    }

//
   return(1);
  }

 hnum = 0;
 while(cnt > 0)
  {
   int hcnt;				//
   if(cnt > hash) { hcnt = hash;  }
   else    	  { hcnt = cnt;   }    
   cnt -= hcnt;				// remain half word copy
   
   while(hcnt--)
    {
     if(*DstAddr != *SrcAddr)
      {
       Print("\nMemory Byte Test Error") ;
       Print("\nSource Data is %02x at Addr %08x",*SrcAddr,SrcAddr) ;
       Print("\nDestin Data is %02x at Addr %08x",*DstAddr,DstAddr) ;
       return(0);
      }
     DstAddr++;
     SrcAddr++;
    }

//
   PrintRotSlash(hnum++);		// print progress indicator
   if(kbd_hit())
    {
     U8 it;
     it = get_byte();
     if(it == KEY_ESC)
      {
       return(-1);			// bcopy skiped
      } 
    }
  }

 return(1);
}

////////////////////////////////////////////////////////////////////////////////
// Half Word Compare Function //////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////


int scomp(void *src, void *dst, int words, int hash)
{
 U16 *SrcAddr;
 U16 *DstAddr;
 int  cnt;
 int hnum;

 SrcAddr = (U16 *)src;			/* half word pointer to source */
 DstAddr = (U16 *)dst;			/* half word pointer to destination */
 cnt	 = words * 2;			/* half word count */

 if(hash <= 0)
  {
// no print prgress indicator //////////////////////////////////////////////////

   while(cnt--)
    {
     if( *DstAddr != *SrcAddr )
      {
       Print("\nMemory Short Test Error") ;
       Print("\nSource Data is %04x at Addr %08x",*SrcAddr,SrcAddr);
       Print("\nDestin Data is %04x at Addr %08x",*DstAddr,DstAddr);
       return(0);
      }
     DstAddr++;				
     SrcAddr++;
    }
//
   return(1);    
  }


 hnum = 0;
 while(cnt > 0)
  {
   int hcnt;				//
   if(cnt > hash) { hcnt = hash; }	//
   else           { hcnt = cnt; }    	//
   cnt -= hcnt;				// remain half word copy
   
   while(hcnt--)
    {
     if( *DstAddr != *SrcAddr)
      {
       Print("\nMemory Short Test Error") ;
       Print("\nSource Data is %04x at Addr %08x",*SrcAddr,SrcAddr) ;
       Print("\nDestin Data is %04x at Addr %08x",*DstAddr,DstAddr) ;
       return(0);
      }
     DstAddr++;
     SrcAddr++;
    }

//
   PrintRotSlash(hnum++);		// print progress indicator
   if(kbd_hit())
    {
     U8 it;
     it = get_byte();
     if(it == KEY_ESC)
      {
       return(-1);			// bcopy skiped
      } 
    }
  }

 return(1);
}

////////////////////////////////////////////////////////////////////////////////
// Word Compare Function ///////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////


int wcomp(void *src, void *dst, int words, int hash)
{
 U32 *SrcAddr;
 U32 *DstAddr;
 int cnt ;
 int hnum;

 SrcAddr = (U32 *) src;				/* word pointer to source */
 DstAddr = (U32 *) dst;				/* word pointer to destination */
 cnt	 = words;				/* word count */

 if(hash == 0)
  {
// no print progress indicator /////////////////////////////////////////////////

   while(cnt--)
    {
     if ( *DstAddr != *SrcAddr )
      {
       Print("\nMemory Long Test Error") ;
       Print("\nSource Data is %08x at Addr %08x",*SrcAddr, SrcAddr);
       Print("\nDestin Data is %08x at Addr %08x",*DstAddr, DstAddr);
       return(0);
      }
     DstAddr++;
     SrcAddr++;
    }

//
   return(1);    
  }

 hnum = 0;
 while(cnt)
  {
   int hcnt;				//
   if(cnt > hash) { hcnt = hash; }
   else           { hcnt = cnt; }    
   cnt -= hcnt;				// remain half word copy
   
   while(hcnt--)
    {
     if ( *DstAddr != *SrcAddr )
      {
       Print("\nMemory Long Test Error") ;
       Print("\nSource Data is %08x at Addr %08x",*SrcAddr, SrcAddr) ;
       Print("\nDestin Data is %08x at Addr %08x",*DstAddr, DstAddr) ;
       return(0);
      }
     DstAddr++;
     SrcAddr++;
    }

//
   PrintRotSlash(hnum++);		// print progress indicator
   if(kbd_hit())
    {
     U8 it;
     it = get_byte();
     if(it == KEY_ESC)
      {
       return(-1);			// bcopy skiped
      } 
    }
  }

// 
 return(1);
}


int PatternGen(unsigned int seed)
{
#ifdef NO_HAVE_RAND
 static int rand_todo = 0;
 if((seed%0x100) == 0) {
    rand_todo = seed;
 }
 seed = (((rand_todo << 11) | (rand_todo >> 21)) + seed);  
 rand_todo = ~seed + 1;  
 return(rand_todo); 
#else 
 if((seed%0x100) == 0) {
    srand(seed);
 }
// 
 return(rand());
#endif 
}

#ifdef NO_HAVE_MEMCPY
void memcpy(void * dst, void *src, int cnt) 
{
 int pos;
 for(pos = 0; pos < cnt; ++pos) {
    *((char *) dst + pos) = *((char *) src + pos);
 }
}
#endif 

/* Initialize Memory Test, Test Pattern Generation */
/* 	src = source address */
/* 	tsize = memory size in words */


static int MemTestInit(U32 *src, int words, int hash)
{
 U32 * addr;
 int cnt;
 int pat;
 int hnum;

 addr = src;				/* source address */
 cnt  = words;				/* source size in words */ 
 pat  = 0;				/* start parameter for PatternGen */

 if(hash <= 0)
  {
// no pring progress indicator /////////////////////////////////////////////////

   while(cnt--) 
    {
     *addr++ = (U32) PatternGen(pat++);
    }
//
   return(1);     
  }

//
 hnum = 0;
 while(cnt) 
  {
   int hcnt;
   if(cnt > hash) { hcnt = hash; }	// 
   else           { hcnt = cnt; } 	// 
   cnt -= hcnt;      			// remain fill words

   while(hcnt--) 
    {
     * addr++ = (U32) PatternGen(pat++);
    }

//
   PrintRotSlash(hnum++);		// print fill progress
   if(kbd_hit())
    {
     U8 it;
     it = get_byte();
     if(it == KEY_ESC)
      {
       return(-1);			// bcopy skiped
      } 
    }
  }

//
 return(1);     
}


////////////////////////////////////////////////////////////////////////////////
// Print Memory Test Status ////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

void PrintRotSlash(int j)
{
 static int i = -1;

 j &= 0x3;
 if(i != j)
  {
   i = j;
        if (j == 0) put_byte('/');
   else if (j == 1) put_byte('-');
   else if (j == 2) put_byte('\\');
   else             put_byte('|');
   put_byte('\b');
  }
}

////////////////////////////////////////////////////////////////////////////////
// Memory  word, halfword, byte copy & Compare test ////////////////////////////
////////////////////////////////////////////////////////////////////////////////


int MemTest(U32 * src, U32 * dst, int tsize, int lsize, int hash)
{
 int rvl; 
 int i;
 int j;

/* Word Test */

 Print("\n  Memory Test (Long)....");
 for(i = 0; i < lsize; ++i) 			/* loop */
  {
   for(j = 0; j < 4; ++j)
    {
// init source area ////////////////////////////////////////////////////////////
     if(MemTestInit(src, tsize, hash) == -1)
      {			
       return(-1);				// ESC pressed 
      }

// word copy ///////////////////////////////////////////////////////////////////
     if(wcopy(src, dst, tsize, hash) == -1)
      {						// skipped test 
       return(-1);				// ESC pressed
      } 

// word compare ////////////////////////////////////////////////////////////////
     rvl = wcomp(src, dst, tsize, hash);
     if(rvl == -1)
      {						// compare skipped
       return(-1);				// ESC pressed
      }
     else
     if(rvl == 0)		
      { 
       Print("Fail.");				/* word test failure */ 
       return(0);
      }
// 
     if(hash > 0)
      {
       break;
      }
     else
      {  
       PrintRotSlash(j++);
      }
    }
  }
 Print("Ok.") ;

/* Half Word Test */

 Print("\n  Memory Test (Short) ..") ;
 for (i=0; i < lsize; i++) 
  {
   for(j = 0; j < 4; ++j)
    {
// init source area ////////////////////////////////////////////////////////////
     if(MemTestInit(src, tsize, hash) == -1)
      {						// init source area
       return(-1);				// skipped
      } 

// half word copy soure to destination /////////////////////////////////////////
     if(scopy(src, dst, tsize, hash) == -1)
      {
       return(-1);				// ESC pressed 
      }

// half word compare source and destination ////////////////////////////////////
     rvl = scomp(src, dst, tsize, hash);
     if(rvl == -1)
      {
       return(-1); 				// ESC pressed
      }
     else
     if(rvl == 0)		
      { 
       Print("Fail.");
       return(0);
      }

// 
     if(hash > 0)
      {
       break;
      }
     else
      { 
       PrintRotSlash(j);
      }
    }
  }
 Print("Ok.");

/* Byte Test */
 Print("\n  Memory Test (Byte) ...") ;
 for (i=0 ; i < lsize ; i++) 
  {
   for(j = 0; j < 4; ++j)
    {
// init source area ////////////////////////////////////////////////////////////
     if(MemTestInit(src, tsize, hash) == -1)
      {
       return(-1);
      } 

// byte copy source to destination /////////////////////////////////////////////
     if(bcopy(src, dst, tsize, hash) == -1)
      {
       return(-1);
      } 

// byte compare source and destination /////////////////////////////////////////
     rvl = bcomp(src, dst, tsize, hash); 
     if(rvl == -1)		
      {
       return(-1); 
      } 
     else
     if(rvl == 0)
      {
       Print("Fail.");
       return(0);
      }

// 
     if(hash > 0)
      {
       break;
      }
     else
      {   
       PrintRotSlash(j);
      }
    }
  }

// 
 Print("Ok") ;
 return(1);
}


////////////////////////////////////////////////////////////////////////////////
// Memory Copy Test Top Program ////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////


static void MemoryCopyTest(void) 
{
 U32 MemCopySrc; 				// memory test source address
 U32 MemCopyDest; 				// memory test destination address
 U32 MemTestSize; 				// memory test size
 U32 *src;    					// memory source address
 U32 *dst;    					// memory destination address
 int tsize;   					// memory Test size
 int lsize;   					// memory test loop size

 Print("\n\nMemory Copy Test");

// memory copy test source location ////////////////////////////////////////////

 Print("\nInput Memory Test Source Location (0x01300000)    : 0x");
 MemCopySrc = get_number(16,0);
 if(MemCopySrc == 0) 
  {						/* default source */
   MemCopySrc = 0x1300000;
  }

// memory copy test destination location ///////////////////////////////////////

 Print("\nInput Memory Test Destination Location(0x01400000): 0x");
 MemCopyDest = get_number(16,0);
 if(MemCopyDest == 0) 
  {						/* default destination */
   MemCopyDest=0x1400000;
  }

// memory copy test size ///////////////////////////////////////////////////////
  
 Print("\nInput Memory Test Size (word) (0x10000)           : 0x");
 MemTestSize = get_number(16,0);
 if(MemTestSize==0) 
  {
   MemTestSize=0x10000;				/* default test size */
  }

// memory copy test repeat count ///////////////////////////////////////////////

 Print("\nInput Memory Test repeat number(0x00005)          : 0x");
 lsize = get_number(16,0);
 if(lsize==0) 
  {				
   lsize= 5; 				/* default repeat count */
  }

 src = (U32 *) MemCopySrc;
 dst = (U32 *) MemCopyDest;
 tsize = (int) MemTestSize;
        
 Print("\nSource Memory Location      : 0x%08x",src);
 Print("\nDestination Memory Location : 0x%08x",dst);
 Print("\nMemory Test Size (words)    : 0x%08x",tsize);
 Print("\nMemory Test Loop Count      : %d\n"  ,lsize);

 if( 
  !MemTest
   (
    src		,		/* memory source address */
    dst		,		/* memory destination address */
    tsize	,		/* test size */
    lsize	,		/* loop count */
    0
   )
  ) 
  { 
   Print("\nFail.\n");
  }
 else 
  {
   Print("\nOk.\n");
  }
}

////////////////////////////////////////////////////////////////////////////////
// External I/O Bank Test //////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////


static void ExtIOBankTest(void)
{
 U32 * base0;			// External Bank 0 Base Address
 U32 * base1;			// External Bank 1 Base Address
 U32 * base2;			// External Bank 2 Base address
 U32 * base3;			// External bank 3 base Address
 U32 testSize; 			// -test size (words)
 int lsize = 5; 		// -Ext IO test loop count

 U32 * src;			// 
 U32 val;			// 
 U8 it;				// -bank num
 U8 cfg;			// 
 int i;				// 

// 
 U32 TCOS0;		 	// -Bank 0 Chip selection set-up time on nOE
 U32 TCOS1;		 	// -Bank 1 Chip selection set-up time on nOE
 U32 TCOS2;		 	// -Bank 2 Chip selection set-up time on nOE
 U32 TCOS3;			// -Bank 3 Chip selection set-up time on nOE

// Address set-up time before nECS (tACS0, tACS1, tACS2, tACS3)	////////////////
 U32 TACS0;			// -Bank 0 Address set-up time before nECS
 U32 TACS1;			// -Bank 1 Address set-up time before nECS
 U32 TACS2;			// -Bank 2 Address set-up time before nECS
 U32 TACS3;			// -Bank 3 Address set-up time before nECS

// Chip selection hold time on nOE (tCOH0,tCOH1,tCOH2,tCOH3) ///////////////////
 U32 TCOH0; 			// -Bank 0 Chip selection hold time on nOE 
 U32 TCOH1;			// -Bank 1 Chip selection hold time on nOE 
 U32 TCOH2;			// -Bank 2 Chip selection hold time on nOE 
 U32 TCOH3;			// -Bank 3 Chip selection hold time on nOE 

 U32 TACC0;			// -Bank 0 Access cycles (nOE Low time)
 U32 TACC1; 			// -Bank 1 Access cycles (nOE Low time)
 U32 TACC2; 			// -Bank 2 Access cycles (nOE Low time)
 U32 TACC3;			// -Bank 3 Access cycles (nOE Low time)

// 
 Print("\nExternal I/O Bank Test");

/* REFEXTCON = DRAM Refresh and External I/O control register */
/* [9:0] External I/O Bank 0 Base pointer */

// External I/O bank 0 base pointer ////////////////////////////////////////////
 base0 = (U32 *) ((REFEXTCON & 0x3FF) << 16 ); 	

 base1 = base0 + 4096;		// External I/O bank 1 base pointer
 base2 = base1 + 4096;		// External I/O bank 2 base pointer
 base3 = base2 + 4096;		// External I/O bank 3 base pointer

 Print("\n  External I/O Bank #0 Base Addr : 0x%08x", base0);
 Print("\n  External I/O Bank #1 Base Addr : 0x%08x", base1);
 Print("\n  External I/O Bank #2 Base Addr : 0x%08x", base2);
 Print("\n  External I/O Bank #3 Base Addr : 0x%08x", base3);

// test source address /////////////////////////////////////////////////////////
 src = (U32 *) 0x01300000;
 Print("\nTest External I/O Bank source address (0x%08x)_0x", src);
 val = get_number(16,0);
 if(val != 0)
  {
   src = (U32 *) val; 
  }

// test size ///////////////////////////////////////////////////////////////////
 testSize = 0x1000; 			/* size in words (4K words)*/
 Print("\nTest External I/O Bank size (words) (0x%08x)  _0x", testSize);
 val = get_number(16,0);
 if(val != 0)
  {
   testSize = val; 
  }

// 
 Print("\nConfigure External I/O Bank Bus Width      [Y/N]_");
 cfg = get_upper();
 if(cfg == 'Y')
  {
////////////////////////////////////////////////////////////////////////////////
// configure external banks bus width //////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

   U32 extbdwth;		// bus width register value 
   U32 dsx0; 			// Data bus width for external I/O bank 0
   U32 dsx1; 			// Data bus width for external I/O bank 1
   U32 dsx2; 			// Data bus width for external I/O bank 2
   U32 dsx3; 			// Data bus width for external I/O bank 3

// 
   extbdwth = EXTDBWTH;		// bus width register value 

/*[21:20] Data bus width for external I/O bank 0 (DSX0)	*/
/*[23:22] Data bus width for external I/O bank 1 (DSX1)	*/
/*[25:24] Data bus width for external I/O bank 2 (DSX2)	*/
/*[27:26] Data bus width for external I/O bank 3 (DSX3)	*/  

// extract banks bus width /////////////////////////////////////////////////////

   dsx0 = (extbdwth>>20) & 0x3; // Data bus width for external I/O bank 0
   dsx1 = (extbdwth>>22) & 0x3; // Data bus width for external I/O bank 1
   dsx2 = (extbdwth>>24) & 0x3; // Data bus width for external I/O bank 2
   dsx3 = (extbdwth>>26) & 0x3; // Data bus width for external I/O bank 3

////////////////////////////////////////////////////////////////////////////////
// read from console bus width for bank 0 //////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

   switch(dsx0)			 	/* Data bus width for external I/O bank 0 */
    {	
     case  0 : dsx0 = 'D'; break; 	// disable 
     case  1 : dsx0 = 'B'; break; 	// byte (8 bit)
     case  2 : dsx0 = 'H'; break; 	// half word (16 bit)
     case  3 : dsx0 = 'W'; break; 	// word (32 bit)
     default : dsx0 = 'D'; break; 
    }

   do
    {
     Print
      (
       "\nExternal I/O Bank 0 Bus Width "
       "[D]isable/[B]yte/[H]alf word/[W]ord (%c)_", 
        dsx0
      );

// 
     cfg = get_upper();			// read upper character  
     switch(cfg)			// convert character to numeric value 
      {
       case 'D' : dsx0 = 0; break; 	// Disable bank 0  
       case 'B' : dsx0 = 1; break; 	// Bank 0 bus width is byte
       case 'H' : dsx0 = 2; break; 	// Bank 0 bus width is half word
       case 'W' : dsx0 = 3; break; 	// Bank 0 bus width is word
       default  : cfg  = 0; break;
      }
    } 
   while(cfg == 0);

////////////////////////////////////////////////////////////////////////////////
// read from console bus width for bank 1 //////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

   switch(dsx1)			 	// conver numeric value to letter
    {
     case  0 : dsx1 = 'D'; break; 	// Disable bank 1
     case  1 : dsx1 = 'B'; break; 	// Bank 1 bus width is byte 
     case  2 : dsx1 = 'H'; break; 	// Bank 1 bus width is half word
     case  3 : dsx1 = 'W'; break; 	// Bank 1 bus width is word
     default : dsx1 = 'D'; break; 	// Disable bank 1
    }

   do
    {
// 
     Print
      (
       "\nExternal I/O Bank 1 Bus Width "
       "[D]isable/[B]yte/[H]alf word/[W]ord (%c)_", 
        dsx1
      );

     cfg = get_upper();			// read character from console 
     switch(cfg)			// conver character to numeric value 	
      {
       case 'D' : dsx1 = 0; break; 	// Disable bank 1 
       case 'B' : dsx1 = 1; break; 	// Bank 1 bus width is byte
       case 'H' : dsx1 = 2; break; 	// Bank 1 bus width is half word
       case 'W' : dsx1 = 3; break; 	// Bank 1 bus width is word
       default  : cfg  = 0; break;	// 	
      }
    } 
   while(cfg == 0);

////////////////////////////////////////////////////////////////////////////////
// read bank 2 bus width from console ///////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

   switch(dsx2)			 	// convert numerical value to letter
    {
     case  0 : dsx2 = 'D'; break; 	// Disable Bank 2	
     case  1 : dsx2 = 'B'; break; 	// Bank 2 bus width is byte
     case  2 : dsx2 = 'H'; break; 	// Bank 2 bus width is half byte
     case  3 : dsx2 = 'W'; break; 	// Bank 2 bus width is word
     default : dsx2 = 'D'; break; 	// Disable Bank 2	
    }

   do
    {
     Print
      (
       "\nExternal I/O Bank 2 Bus Width "
       "[D]isable/[B]yte/[H]alf word/[W]ord (%c)_", 
        dsx2
      );

// 
     cfg = get_upper();			// read character from console 
     switch(cfg)			// convert character to numeriacal value 
      {
       case 'D' : dsx2 = 0; break; 	// Disable bank 
       case 'B' : dsx2 = 1; break; 	// Bank 2 bus size is byte
       case 'H' : dsx2 = 2; break; 	// Bank 2 bus size is half word 
       case 'W' : dsx2 = 3; break; 	// Bank 2 bus size is word
       default  : cfg  = 0; break;
      }
    } 
   while(cfg == 0);

////////////////////////////////////////////////////////////////////////////////
// read bank 3 bus width from console //////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

   switch(dsx3)			 	/* Data bus width for external I/O bank 0 */
    {
     case  0 : dsx3 = 'D'; break; 	// Disable Bank 3
     case  1 : dsx3 = 'B'; break; 	// Bank 3 bus size is byte 
     case  2 : dsx3 = 'H'; break; 	// Bank 3 bus size is half word
     case  3 : dsx3 = 'W'; break; 	// Bank 3 bus size is word
     default : dsx3 = 'D'; break; 	// Disable bank 3
    }

   do
    {
     Print
      (
       "\nExternal I/O Bank 3 Bus Width "
       "[D]isable/[B]yte/[H]alf word/[W]ord (%c)_", 
        dsx3
      );
// 
     cfg = get_upper();
     switch(cfg)			/* Data bus width for external I/O bank 0 */
      {
       case 'D' : dsx3 = 0; break; 
       case 'B' : dsx3 = 1; break; 
       case 'H' : dsx3 = 2; break; 
       case 'W' : dsx3 = 3; break; 
       default  : cfg = 0; break;
      }
    } 
   while(cfg == 0);

// 
   extbdwth &= ~EXTDBWTH_DSX0; 		/* Data bus width for external I/O bank 0 */
   extbdwth &= ~EXTDBWTH_DSX1; 		/* Data bus width for external I/O bank 1 */
   extbdwth &= ~EXTDBWTH_DSX2; 		/* Data bus width for external I/O bank 2 */
   extbdwth &= ~EXTDBWTH_DSX3; 		/* Data bus width for external I/O bank 3 */

//
   extbdwth      |=

    (dsx0 << 20) |
    (dsx1 << 22) |
    (dsx2 << 24) |
    (dsx3 << 26);

   Print("\nEXTDBWTH = 0x%08x",extbdwth);
   EXTDBWTH = extbdwth; 		/* Data bus width for external I/O bank 0 */
  } 

////////////////////////////////////////////////////////////////////////////////
// read tested bank num from console ///////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

 Print("\nSelect Test External I/O Bank  (0, 1, 2, 3, Q)  _");
 it = get_upper();
 if(it >= '0' && it <= '3')
  {
   it = it - '0';			// Bank 0 
  }
 else
 if(it == 'Q')
  {
   return; 
  }
 else
  {
   Print("\n  Invalid External I/O Bank Selected [%c]", it);
   return;
  } 	


////////////////////////////////////////////////////////////////////////////////
// configure external I/O bank /////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

 Print("\nConfigure External I/O Bank %c [Y]es/[N]o/[D]efault/[A]ll/[Q]uit]? ", it);
 cfg = get_upper();
 if(cfg == 'Q')
  {
// Quit ////////////////////////////////////////////////////////////////////////
   return;
  }

 if(cfg == 'D')
  {
// default configuration // [D]efault //////////////////////////////////////////

/* Chip selection set-up time on nOE (tCOS0, tCOS1, tCOS2, tCOS3) */
/* [0:2]   EXTACON0 tCOS0 */
/* [0:2]   EXTACON1 tCOS2 */
/* [18:16] EXTACON0 tCOS1 */
/* [18:16] EXTACON1 tCOS3 */

   TCOS0 = (0x7 << EXTACON0_TCOS0_OFFSET); // Bank 0 Chip selection set-up time on nOE
   TCOS1 = (0x7 << EXTACON0_TCOS1_OFFSET); // Bank 1 Chip selection set-up time on nOE
   TCOS2 = (0x7 << EXTACON1_TCOS2_OFFSET); // Bank 2 Chip selection set-up time on nOE
   TCOS3 = (0x7 << EXTACON1_TCOS3_OFFSET); // Bank 3 Chip selection set-up time on nOE

/* Address set-up time before nECS (tACS0, tACS1, tACS2, tACS3)	*/ 
/* [5:3]   EXTACON0 tACS0 */
/* [5:3]   EXTACON1 tACS2 */
/* [21:19] EXTACON0 tACS1 */
/* [21:19] EXTACON1 tACS3 */
   TACS0 = (0x6 << EXTACON0_TACS0_OFFSET); // Bank 0 Address set-up time before nECS
   TACS1 = (0x6 << EXTACON0_TACS1_OFFSET); // Bank 1 Address set-up time before nECS
   TACS2 = (0x6 << EXTACON1_TACS2_OFFSET); // Bank 2 Address set-up time before nECS
   TACS3 = (0x6 << EXTACON1_TACS3_OFFSET); // Bank 3 Address set-up time before nECS

// Chip selection hold time on nOE (tCOH0,tCOH1,tCOH2,tCOH3) ///////////////////
/* [8:6]   EXTACON0 tCOH0 */
/* [8:6]   EXTACON1 tCOH2 */
/* [24:22] EXTACON0 tCOH1 */
/* [24:22] EXTACON1 tCOH3 */
   TCOH0 = (0x6 << EXTACON0_TCOH0_OFFSET); // Bank 0 Chip selection hold time on nOE 
   TCOH1 = (0x6 << EXTACON0_TCOH1_OFFSET); // Bank 1 Chip selection hold time on nOE 
   TCOH2 = (0x6 << EXTACON1_TCOH2_OFFSET); // Bank 2 Chip selection hold time on nOE 
   TCOH3 = (0x6 << EXTACON1_TCOH3_OFFSET); // Bank 3 Chip selection hold time on nOE 

// Access cycles (nOE low time (tACC0,tACC1,tACC2,tACC3)
// [11:9]  EXTACON0 tACC0 
// [11:9]  EXTACON1 tACC2 
// [27:25] EXTACON0 tACC1 
// [27:25] EXTACON1 tACC3 
   TACC0 = (0x7 << EXTACON0_TACC0_OFFSET); // Bank 0 Access cycles (nOE Low time)
   TACC1 = (0x7 << EXTACON0_TACC1_OFFSET); // Bank 1 Access cycles (nOE Low time)
   TACC2 = (0x7 << EXTACON1_TACC2_OFFSET); // Bank 2 Access cycles (nOE Low time)
   TACC3 = (0x7 << EXTACON1_TACC3_OFFSET); // Bank 3 Access cycles (nOE Low time)
  }
 else
 if(cfg != 'N')
  {
////////////////////////////////////////////////////////////////////////////////
// [Y]es / [A]ll ///////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

// Chip selection set-up time on nOE (tCOS0, tCOS1, tCOS2, tCOS3) //////////////
   TCOS0 = (EXTACON0 >> EXTACON0_TCOS0_OFFSET) & 0x7; // Bank 0 Chip selection set-up time on nOE
   TCOS1 = (EXTACON0 >> EXTACON0_TCOS1_OFFSET) & 0x7; // Bank 1 Chip selection set-up time on nOE
   TCOS2 = (EXTACON1 >> EXTACON1_TCOS2_OFFSET) & 0x7; // Bank 2 Chip selection set-up time on nOE
   TCOS3 = (EXTACON1 >> EXTACON1_TCOS3_OFFSET) & 0x7; // Bank 3 Chip selection set-up time on nOE

// Address set-up time before nECS (tACS0, tACS1, tACS2, tACS3)	////////////////
   TACS0 = (EXTACON0 >> EXTACON0_TACS0_OFFSET) & 0x7; // Bank 0 Address set-up time before nECS
   TACS1 = (EXTACON0 >> EXTACON0_TACS1_OFFSET) & 0x7; // Bank 1 Address set-up time before nECS
   TACS2 = (EXTACON1 >> EXTACON1_TACS2_OFFSET) & 0x7; // Bank 2 Address set-up time before nECS
   TACS3 = (EXTACON1 >> EXTACON1_TACS3_OFFSET) & 0x7; // Bank 3 Address set-up time before nECS

// Chip selection hold time on nOE (tCOH0,tCOH1,tCOH2,tCOH3) ///////////////////
   TCOH0 = (EXTACON0 >> EXTACON0_TCOH0_OFFSET) & 0x7; // Bank 0 Chip selection hold time on nOE 
   TCOH1 = (EXTACON0 >> EXTACON0_TCOH1_OFFSET) & 0x7; // Bank 1 Chip selection hold time on nOE 
   TCOH2 = (EXTACON1 >> EXTACON1_TCOH2_OFFSET) & 0x7; // Bank 2 Chip selection hold time on nOE 
   TCOH3 = (EXTACON1 >> EXTACON1_TCOH3_OFFSET) & 0x7; // Bank 3 Chip selection hold time on nOE 

// Access cycles (nOE low time (tACC0,tACC1,tACC2,tACC3) ///////////////////////
   TACC0 = (EXTACON0 >> EXTACON0_TACC0_OFFSET) & 0x7; // Bank 0 Access cycles (nOE Low time)
   TACC1 = (EXTACON0 >> EXTACON0_TACC1_OFFSET) & 0x7; // Bank 1 Access cycles (nOE Low time)
   TACC2 = (EXTACON1 >> EXTACON1_TACC2_OFFSET) & 0x7; // Bank 2 Access cycles (nOE Low time)
   TACC3 = (EXTACON1 >> EXTACON1_TACC3_OFFSET) & 0x7; // Bank 3 Access cycles (nOE Low time)

   for(i = 0; i < 4; )
    {
     U32 TCOS; 			// Chip selection set-up time on nOE
     U32 TACS; 			// Address set-up time before nECS
     U32 TCOH; 			// Chip selection hold time on nOE 
     U32 TACC; 			// Access cycles (nOE Low time)

     if(cfg == 'Y' && it != i)
      {
       i += 1; 
       continue;
      }

     while(1)
      {
// Chip selection set-up time on nOE ///////////////////////////////////////////
       Print("\nBank %d Chip selection set-up time on nOE tCOS%d (0-7)_",i,i);
       TCOS = get_upper(); 			
       if(TCOS >= '0' && TCOS <= '7')
        {
         TCOS -= '0';		// 
         break;
        }
       else
        {
         Print("\nInvalid tCOS%d %c", i, TCOS);
         continue;
        }
      }

// Address set-up time before nECS tACS ////////////////////////////////////////
 
     while(1)
      {
       Print("\nBank %d Address set-up time before nECS tACS%d   (0-7)_",i,i);
       TACS = get_upper(); 			
       if(TACS >= '0' && TACS <= '7')
        {
         TACS -= '0';
         break;
        }
       else
        {
         Print("\nInvalid tACS%d %c", i, TACS);
         continue;
        }
      }

// Chip selection hold time on nOE tCOH ////////////////////////////////////////

     while(1)
      {
       Print("\nBank %d Chip selection hold time on nOE tCOH%d (0-7)  _",i,i);
       TCOH = get_upper();
       if(TCOH >= '0' && TCOH <= '7')
        {
         TCOH -= '0';
         break;
        }
       else
        {
         Print("\nInvalid tCOH%d %c", i, TACS);
         continue;
        }
      }

     while(1)
      {
       Print("\nBank %d Access cycles (nOE low time) tACC%d (0-7)     _",i,i);
       TACC = get_upper(); 			// Chip selection set-up time on nOE
       if(TACC >= '0' && TACC <= '7')
        {
         TACC -= '0';
         break; 
        }
       else
        {
         Print("\nInvalid tCOH%d %c", i, TACC);
         continue;
        }
      }

     if(i == 0)
      {
       TCOS0 = (TCOS << EXTACON0_TCOS0_OFFSET); // Bank 0 Chip selection set-up time on nOE
       TACS0 = (TACS << EXTACON0_TACS0_OFFSET); // Bank 0 Address set-up time before nECS
       TCOH0 = (TCOH << EXTACON0_TCOH0_OFFSET); // Bank 0 Chip selection hold time on nOE 
       TACC0 = (TACC << EXTACON0_TACC0_OFFSET); // Bank 0 Access cycles (nOE Low time)
      } 
     else 
     if(i == 1)
      {
       TCOS1 = (TCOS << EXTACON0_TCOS1_OFFSET); // Bank 1 Chip selection set-up time on nOE
       TACS1 = (TACS << EXTACON0_TACS1_OFFSET); // Bank 1 Address set-up time before nECS
       TCOH1 = (TCOH << EXTACON0_TCOH1_OFFSET); // Bank 1 Chip selection hold time on nOE 
       TACC1 = (TACC << EXTACON0_TACC1_OFFSET); // Bank 1 Access cycles (nOE Low time)
      }
     else
     if(i == 2)
      {
       TCOS2 = (TCOS << EXTACON1_TCOS2_OFFSET); // Bank 2 Chip selection set-up time on nOE
       TACS2 = (TACS << EXTACON1_TACS2_OFFSET); // Bank 2 Address set-up time before nECS
       TCOH2 = (TCOH << EXTACON1_TCOH2_OFFSET); // Bank 2 Chip selection hold time on nOE 
       TACC2 = (TACC << EXTACON1_TACC2_OFFSET); // Bank 2 Access cycles (nOE Low time)
      }
     else
     if(i == 3)
      {
       TCOS3 = (TCOS << EXTACON1_TCOS3_OFFSET); // Bank 3 Chip selection set-up time on nOE
       TACS3 = (TACS << EXTACON1_TACS3_OFFSET); // Bank 3 Address set-up time before nECS
       TCOH3 = (TCOH << EXTACON1_TCOH3_OFFSET); // Bank 3 Chip selection hold time on nOE 
       TACC3 = (TACC << EXTACON1_TACC3_OFFSET); // Bank 3 Access cycles (nOE Low time)
      }
//
     i += 1;
    }
  } 
  
 if(
   cfg == 'Y' ||			// Yes
   cfg == 'D' ||			// Default 
   cfg == 'A'				// All 
  ) 
  {
  EXTACON0  = 
// 
   TCOS0 | 
   TACS0 | 
   TCOH0 | 
   TACC0 | 
// 
   TCOS1 | 
   TACS1 | 
   TCOH1 | 
   TACC1; 

  EXTACON1  = 
// 
   TCOS2 | 
   TACS2 | 
   TCOH2 | 
   TACC2 | 
// 
   TCOS3 | 
   TACS3 | 
   TCOH3 | 
   TACC3; 
 }

 if(it == 0)
  {
   Print("\nExternal I/O Bank #0 Test ");
   MemTest(src, base0, testSize, lsize, 0);
  }
 else 
 if(it == 1)
  {
   Print("\nExternal I/O Bank #1 Test ") ;
   MemTest(src, base1, testSize,lsize, 0);
  }
 else 
 if(it == 2)
  {
   Print("\nExternal I/O Bank #2 Test ") ;
   MemTest(src, base2, testSize,lsize, 0);
  }
 else 
 if(it == 3)
  {
   Print("\nExternal I/O Bank #3 Test ") ;
   MemTest(src, base3, testSize,lsize, 0);
  }
 else 
 Print("\nInvalid External I/O Bank Selected") ;
}

////////////////////////////////////////////////////////////////////////////////
// Memory Test Menu ////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

void MemoryTest(void) 
{
 char it;				// selected item 
 do 
  {
   Print("\nMemory Test\n") ;
   Print("[V] View Memory Configuration\n");
   Print("[C] Memory Copy Test\n");
   Print("[R] Memory Dump\n");
   Print("[F] Memory Pattern Fill\n");
   Print("[S] Memory Pattern Search\n");
   Print("[E] External I/O Bank Test\n");
   Print("[Q] Quit\n");

   Print("\nSelect Test Item : ");
   it = get_upper();
	
   switch(it)
    {
     case 'V' : MemViewConf();      break; /* View memory configuration */
     case 'C' : MemoryCopyTest();   break; /* test memory, word, half word, byte */
     case 'R' : MemoryDump();       break; /* dump memory */
     case 'F' : MemPatternFill();   break; /* pattern fill */	
     case 'S' : MemPatternSearch(); break; /* pattern search  */
     case 'E' : ExtIOBankTest();    break; /*  */
     case 'Q' : return;
     default  : Print("\nInvalid Test Item Selected") ; 
      break;
    }
   Print("\n Press Any Key To Continue ....") ; 
   it = get_upper();
  } while((it != 'Q'));
}


